"""
Functions to handle communication with an endpoint.
"""
import locale
from pathlib import Path
import socket
import time
import paramiko
from yvpn.util import get_spinner
from rich.prompt import Prompt


def key_exchange(ssh_pubkey_path: str, server_ip: str, client_ip: str) -> str:
    """exchange wireguard keys via ssh"""

    with get_spinner() as spinner:
        spinner.add_task("Waiting for server to come up...")
        while not endpoint_server_up(server_ip):
            time.sleep(1)

    ssh_key = ssh_pubkey_path.replace(".pub", "")
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    with get_spinner() as spinner:
        spinner.add_task("Performing key exchange with new VPN endpoint ...")
        ssh.connect(server_ip, username="root",
                    key_filename=ssh_key,
                    look_for_keys=False,
                    banner_timeout=60,
                    timeout=60,
                    auth_timeout=60)

    # activate client on server
    encoding = locale.getpreferredencoding()
    client_public_key = Path("/etc/wireguard/public.key") \
        .read_text(encoding=encoding).strip()
    command = f"wg set wg0 peer {client_public_key} allowed-ips {client_ip}"
    ssh.exec_command(command)

    # get and return server public key
    _, stdout, _ = ssh.exec_command("cat /etc/wireguard/public.key")
    server_public_key = stdout.read().decode().strip()

    return server_public_key


def endpoint_server_up(server_ip: str) -> bool:
    """Check and see if an ssh server is up"""
    connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        connection.connect((server_ip, 22))
        connection.shutdown(2)
        return True
    except ConnectionRefusedError:
        return False


def prompt_to_fix(name: str) -> str:
    action = Prompt.ask(f"Problem with {name}'s wireguard config, what do you want to do?",
                        choices=["delete", "repair"],
                        default="delete")

    return action



"""
Functions to manage wireguard
"""
import locale
from glob import glob
from pathlib import Path
import subprocess


def refresh_keys(overwrite_existing: bool = False):
    """Check if keys exist and create them, or force them to be overwritten"""

    subprocess.run(["sudo", "chmod", "755", "/etc/wireguard"],
                   check=True)

    keys_exist = Path("/etc/wireguard/private.key").is_file() and \
                 Path("/etc/wireguard/public.key").is_file()

    if not keys_exist or overwrite_existing:

        # unlock /etc/wireguard permissions
        subprocess.run(["sudo", "chmod", "-R", "777", "/etc/wireguard"],
                       check=True)

        # generate and save fresh wireguard private key
        private_key = subprocess.run(["wg", "genkey"], capture_output=True,
                                     check=True).stdout.decode()
        enc = locale.getpreferredencoding()
        with open("/etc/wireguard/private.key", "w", encoding=enc) as key_file:
            key_file.write(private_key)

        # generate and save wireguard public key
        with subprocess.Popen(["cat", "/etc/wireguard/private.key"],
                              stdout=subprocess.PIPE) as private_key:
            public_key = subprocess.check_output(["wg", "pubkey"],
                                                 stdin=private_key.stdout).decode()
        enc = locale.getpreferredencoding()
        with open("/etc/wireguard/public.key", "w", encoding=enc) as key_file:
            key_file.write(public_key)

        # lock /etc/wireguard permissions
        subprocess.run(["sudo", "chmod", "-R", "755", "/etc/wireguard"], check=True)
        subprocess.run(["sudo", "chmod", "700", "/etc/wireguard/private.key"],
                       check=True)


def get_client_private_key() -> str:
    """get the wireguard private key"""
    subprocess.run(["sudo", "chmod", "644", "/etc/wireguard/private.key"],
                   check=True)
    enc = locale.getpreferredencoding()
    with open("/etc/wireguard/private.key", encoding=enc) as file:
        private_key = file.read().strip()
    subprocess.run(["sudo", "chmod", "600", "/etc/wireguard/private.key"],
                   check=True)
    return private_key


def configure_client(endpoint_name: str,
                     server_public_key: str,
                     server_ip: str, client_ip: str) -> None:
    """create the wireguard config"""
    print("Setting up local configuration ...")

    config = ("[Interface]",
              f"PrivateKey = {get_client_private_key()}",
              f"Address = {client_ip}/24",
              "\n",
              "[Peer]",
              f"PublicKey = {server_public_key}",
              f"Endpoint = {server_ip}:51820",
              "AllowedIPs = 0.0.0.0/0",
              "\n"
              )

    config_file = f"/etc/wireguard/{endpoint_name}.conf"
    subprocess.run(["sudo", "touch", config_file], check=True)
    subprocess.run(["sudo", "chmod", "666", config_file], check=True)
    enc = locale.getpreferredencoding()
    with open(config_file, "w", encoding=enc) as file:
        file.write("\n".join(config))
    subprocess.run(["sudo", "chmod", "600", config_file], check=True)


def config_exists(name: str) -> bool:
    interfaces = [x.split("/")[-1] for x in glob("/etc/wireguard/*.conf")]
    return name + ".conf" in interfaces


def wireguard_installed() -> bool:
    """Check to see if wireguard is installed"""
    sp = subprocess.run(["which", "wg"], capture_output=True)
    return True if sp.returncode == 0 else False



"""
Utility functions
"""
import locale
import subprocess
from os import path
from pathlib import Path
from typing import Tuple
from rich.progress import Progress, SpinnerColumn, TextColumn


def get_ssh_pubkey() -> Tuple[str, str]:
    """get and return the ssh pubkey text and path"""
    key_path = path.expanduser("~/.ssh/")
    key_name = "yvpn"
    pubkey_path = f"{key_path}{key_name}.pub"
    encoding = locale.getpreferredencoding()
    try:
        pubkey = Path(pubkey_path).read_text(encoding=encoding).strip()
    except FileNotFoundError:
        subprocess.run(["ssh-keygen", "-f", f"{key_path}{key_name}", "-N", ""],
                       check=True)
        pubkey = Path(pubkey_path).read_text(encoding=encoding).strip()
    return pubkey, pubkey_path


def get_datacenter_name(name: str) -> str:
    """turn a slug into a city name"""
    slug = name[-4:-1]

    slug_to_city = {
        'ams': "Amsterdam",
        'nyc': "New York City",
        'sfo': "San Francisco",
        'sgp': "Singapore",
        'lon': "London",
        'fra': "Frankfurt",
        'tor': "Toronto",
        'blr': "Bangalore"
    }
    return slug_to_city[slug]


def get_spinner():
    """return a progress spinner"""
    spinner = Progress(SpinnerColumn(),
                       TextColumn("{task.description}[progress.description]"),
                       transient=False)
    return spinner



"""
Functions that make requests to the yvpn server.
"""

import requests
from yvpn.config import TOKEN, SERVER_URL


def get_first_endpoint():
    """return the name of a user's first endpoint"""
    header = {"token": f"{TOKEN}"}
    endpoints = requests.get(url=f"{SERVER_URL}/status",
                             headers=header).json()
    return endpoints[0]["endpoint_name"]


def handle_endpoint_name_or_number(user_input: str) -> str:
    """allow a user to select an endpoint by name or number"""
    header = {"token": f"{TOKEN}"}
    endpoints: dict = requests.get(url=f"{SERVER_URL}/status",
                                   headers=header).json()

    # if it's not an appropriate number, it must be a name or user error
    if not user_input.isnumeric() or int(user_input) > len(endpoints):
        return user_input

    return endpoints[int(user_input)]["endpoint_name"]


def get_datacenter_regions() -> list:
    """return a list of available datacenters"""
    print("Getting a list of available datacenters ...")
    header = {"token": f"{TOKEN}"}
    regions = requests.get(url=f"{SERVER_URL}/datacenters",
                           headers=header).json()["available"]

    return regions



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yvpn']

package_data = \
{'': ['*']}

install_requires = \
['paramiko>=2.11.0,<3.0.0',
 'requests>=2.28.1,<3.0.0',
 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['yvpn = yvpn.main:main']}

setup_kwargs = {
    'name': 'yvpn',
    'version': '0.1.1',
    'description': 'Command line tool to manage yvpn endpoints',
    'long_description': '# yvpn\n> A CLI client to manage YOUR vpn.\n\n## Overview\n\nThe yvpn tool is a command line client that allows users to manage VPN endpoints by communicating with the yourvpn.info service.  \n\nUsers can create, use, and destroy VPN endpoints at will.  These endpoints are Virtual Private Servers that are created for the exclusive use of the user unlike other VPN providers where thousands of users share a single endpoint.\n\nThis tool is meant to be extensible.  Beneath the hood SSH is used for the initial wireguard key exchange and future releases will allow users to quickly drop directly into a shell on the remote machine, perform file transfers, or remotely execute scripts.  \n\nThis is not just a a tool for managing your VPN needs, but also a powerful resource for quickly deploying on demand services limited only by your creativity.\n\n## Installation\n\nThe quickest way to get up and running is to install the tool from pip:\n\n`pip3 install yvpn`\n\nYou will need wireguard:\n\n(https://www.wireguard.com/install/)(https://www.wireguard.com/install/)\n\nYou need to set two environment variables:\n\n1. The server url:\n\t`URL_yVPN="https://yourvpn.info"`\n\n2. Your token:\n\t`TOKEN_yVPN="<token>"`\n\n## Where to get tokens\n\nRight now yvpn is in closed alpha.  Tokens are available by invitation only.\n\n## What even is a token?  Like an account?\n\nNo, we do not and will never offer accounts.  Privacy is a core principal and we figure the best way to preserve user\'s privacy is to simply store zero user information.  That\'s where Tokens come in.\n\nThink of a token as a sort of prepaid calling card.  Remember those good old days?  Where you\'d buy a calling card with however many minutes preloaded and then you had to call a 1-800 number from a payphone and enter the little code beneath the scratch off material?  That\'s what our token model will be.  \n\nOne day, once we\'re ready for a beta, there will be a simple storefront where you can buy a token preloaded with some amount of credit and off you go.\n\n## How will billing work then?\n\nThere will be no billing as in no invoicing.  Your token\'s fund balance will be debited based directly on usage.  If you don\'t have any endpoints running, you won\'t pay anything. \n\n## Overview of commands\n\n### `yvpn clean`\n\nDestroys all of your endpoints and refreshes your wireguard keys.\n\n### `yvpn connect`\n\nConnect to an endpoint.  You can pass the name or number of the endpoint to connect to a specific one, i.e. `yvpn connect 3`, or automatically connect to the first one on the list without.\n\n### `yvpn create`\n\nCreate a new endpoint.  You can optionally specify a specific datacenter, `yvpn create lon1`, see `yvpn datacenters` below, or create a new endpoint in a randomly selected datacenter by omitting the final argument.\n\n### `yvpn datacenters`\n\nDisplays a list of the currently available datacenters.  \n\n### `yvpn destroy <num>`\n\nDestroy the specified endpoint.\n\n### `yvpn disconnect`\n\nDisconnect from the active endpoint.\n\n### `yvpn status`\n\nDisplay a table of your endpoints with a number, name, location, and creation timestamp.  Also displays your token\'s balance and expected depletion date at current usage.\n\n',
    'author': 'Ben Simcox',
    'author_email': 'ben@bnsmcx.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



#! /usr/bin/env python3

"""
Homebase for the typer CLI app.  This file contains only functions that
represent commands available to the user.
"""

import subprocess
import sys
from glob import glob
import requests
import typer
from rich.table import Table

from yvpn import util
from yvpn import api_calls
from yvpn import wireguard
from yvpn import endpoint
from yvpn.config import SERVER_URL, TOKEN, console

app = typer.Typer(no_args_is_help=True,
                  add_completion=False)


@app.command()
def create(region: str = typer.Argument("random")):
    """CREATE a new VPN endpoint, random by default"""

    wireguard.refresh_keys()
    ssh_pubkey, ssh_pubkey_path = util.get_ssh_pubkey()

    with util.get_spinner() as spinner:
        spinner.add_task("Creating the endpoint, this could take a minute.")
        header = {"token": f"{TOKEN}"}
        response = requests.post(url=f"{SERVER_URL}/create",
                                 json={'region': f'{region}',
                                       'ssh_pub_key': f'{ssh_pubkey}'},
                                 headers=header)

    if response.status_code != 200:
        console.print(f"There was a problem:\n {response.json()}")
        sys.exit(1)

    endpoint_ip = response.json()["server_ip"]
    endpoint_name = response.json()["endpoint_name"]
    client_ip = "10.0.0.2"
    server_public_key = endpoint.key_exchange(ssh_pubkey_path,
                                              endpoint_ip, client_ip)

    if not server_public_key:
        console.print("Key exchange failed.")
        destroy(endpoint_name)
        sys.exit(1)

    wireguard.configure_client(endpoint_name,
                               server_public_key,
                               endpoint_ip,
                               client_ip)

    console.print("New endpoint successfully created and configured.")


@app.command()
def datacenters():
    """GET a list of available datacenters for endpoint creation"""
    console.print(api_calls.get_datacenter_regions())


@app.command()
def connect(endpoint_name: str = typer.Argument(api_calls.get_first_endpoint)):
    """CONNECT to an endpoint by name or number"""
    endpoint_name = api_calls.handle_endpoint_name_or_number(endpoint_name)
    disconnect()
    command = subprocess.run(["sudo", "wg-quick", "up", endpoint_name],
                             check=True,
                             capture_output=True)

    if not command.returncode == 0:
        console.print(command.stderr)
    console.print(f"[bold green]Connected to {endpoint_name}")


@app.command()
def disconnect():
    """DISCONNECT from the active endpoint"""
    interfaces = glob("/etc/wireguard/*.conf")
    for interface in interfaces:
        try:
            subprocess.run(["sudo", "wg-quick", "down", interface],
                           check=True,
                           capture_output=True)
        except subprocess.CalledProcessError:
            # ignore error when trying to bring down an inactive interface
            continue


@app.command()
def clean():
    """DELETE and REFRESH all keys, DESTROY all endpoints"""
    disconnect()
    interfaces = glob("/etc/wireguard/*.conf")
    wireguard.refresh_keys(True)
    for interface in interfaces:
        destroy(interface.replace('.conf', "").replace("/etc/wireguard/", ""))


@app.command()
def destroy(endpoint_name: str = typer.Argument(api_calls.get_first_endpoint)):
    """permanently DESTROY an endpoint"""

    disconnect()
    endpoint_name = api_calls.handle_endpoint_name_or_number(endpoint_name)
    header = {"token": f"{TOKEN}"}
    deletion_request = requests.delete(url=f"{SERVER_URL}/endpoint",
                                       headers=header,
                                       params={'endpoint_name': f'{endpoint_name}'})

    if deletion_request.status_code == 200:
        try:
            subprocess.run(
                ["sudo", "rm", f"/etc/wireguard/{endpoint_name}.conf"],
                check=True,
                capture_output=True)
            console.print(f"{endpoint_name} successfully deleted.")
        except subprocess.CalledProcessError:
            console.print(f"{endpoint_name} deleted but couldn't delete the wireguard config.")
            console.print("This is probably because it does not exist.")
    else:
        console.print(f"Problem deleting {endpoint_name}:\n {deletion_request.json()}")


@app.command()
def status():
    """display connection, usage and endpoint info"""

    active_connection = subprocess.run(["sudo", "wg", "show"],
                                       capture_output=True,
                                       check=True)

    connection_info = "[bold]Not connected."
    token_info = "[bold]Token will be depleted on TODO at current usage."

    if active_endpoint := active_connection.stdout:
        active_endpoint = active_endpoint.decode().split()[1]
        connection_info = f"[bold green]Connected to: {active_endpoint}"

    endpoint_table = Table(caption=token_info, expand=True)

    endpoint_table.add_column("Number", justify="center")
    endpoint_table.add_column("Name", justify="center")
    endpoint_table.add_column("Location", justify="center")
    endpoint_table.add_column("Created", justify="center")

    header = {"token": f"{TOKEN}"}
    server_status = requests.get(url=f"{SERVER_URL}/status",
                                 headers=header)

    if server_status.status_code != 200:
        console.print("[red bold]There was a problem:")
        console.print(server_status.json())
        sys.exit(1)

    for index, endpoint_server in enumerate(server_status.json()):
        name = endpoint_server["endpoint_name"]
        if not wireguard.config_exists(name):
            match endpoint.prompt_to_fix(name):
                case "delete":
                    destroy(name)
                    sys.exit(1)
                case "repair":
                    print("TODO: Not implemented, you can only 'delete' broken endpoints")
        location = util.get_datacenter_name(name)
        endpoint_style = "bold"
        if active_endpoint == name:
            endpoint_style = "bold green"
        endpoint_table.add_row(str(index), name, location, "TODO",
                               style=endpoint_style)

    console.print()
    console.rule(connection_info)
    console.print(endpoint_table, justify="center")
    console.print()


def main():
    """The entry point into the app."""
    if not wireguard.wireguard_installed():
        console.print(
            "[bold red]Wireguard not found, please install. www.wireguard.com/install/")
        sys.exit(1)
    app()


if __name__ == "__main__":
    main()



"""
Handle setting the two global variables the app needs.
"""
import sys
from os import environ

import requests
from requests.exceptions import InvalidSchema, MissingSchema, InvalidURL, ConnectionError
from rich.console import Console


def get_user_token() -> str:
    """Try to get token from an environment variable, ask the user otherwise"""
    try:
        return environ['TOKEN_yVPN']
    except KeyError:
        token = input("Enter token: ")
        print("Set the 'TOKEN_yVPN' environment variable to skip this in the future.")
        return token


def test_server_connection(url: str):
    try:
        header = {"token": f"{TOKEN}"}
        requests.get(url=f"{url}/status", headers=header)
    except (InvalidURL, InvalidSchema, MissingSchema, ConnectionError) as e:
        print(f"There is a problem with your server url:\n{e}")
        print(f"Ensure it is set correctly to include the http/https prefix.")
        sys.exit(1)


def get_server_url():
    try:
        url = environ['URL_yVPN']
        test_server_connection(url)
        return url
    except KeyError:
        url = input("Enter server url: ")
        test_server_connection(url)
        print("Set the 'URL_yVPN' environment variable to skip this in the future.")
        return url


TOKEN = get_user_token()
SERVER_URL = get_server_url()
console = Console()
